/* ----------------------------------------------------------------------------- */
/*  */
/*  DESCRIPTION: */
/* 	HERETIC graphics stuff for Svgalib, Linux. */
/* 	Keyboard routine inspired from Snes9x - SVGALIB */
/*  */
/* ----------------------------------------------------------------------------- */

#include <stdlib.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <errnos.h>
#include <signal.h>

#include <vga.h>
#include <vgakeyboard.h>

#include "i_system.h"
/* #include "v_video.h" */

#include "h2def.h"
#include "r_local.h"
#include "am_map.h"

/*  bpp = 8 , pixel_size = 1 , mode = 320x200 */
int		bpp;
int		fbnum=0;

int		pixel_size;
lighttable_t	truecolor_palette[256];
extern byte	*orig_colormaps;

static char prev_keystate [128] = "";

extern	unsigned long fuzzmask;	/*  for fuzz effect in 15/16 bit mode */

void *svgascreen[2];

void I_ShutdownGraphics(void)
{
	keyboard_close();

	vga_setmode(TEXT);	
}

/*  */
/*  I_StartTic */
/*  */
void I_StartTic (void)
{
	event_t	event;

	/*  Taken from Snes9x */

	char	*keystate;

	keyboard_update ();
	keystate=keyboard_getstate();

#define KEY_DOWN(a) (keystate[a])
#define KEY_PRESS(a) (keystate[a] && !prev_keystate[a])
#define KEY_WASPRESSED(a) (prev_keystate[a] && !keystate[a])
#define PROCESS_KEY(k,t)					\
	if (KEY_PRESS(k))					\
	{							\
			event.type = ev_keydown;		\
			event.data1 = t;			\
			H2_PostEvent(&event);			\
	}							\
	if (KEY_WASPRESSED(k))					\
	{							\
			event.type = ev_keyup;			\
			event.data1 = t;			\
			H2_PostEvent(&event);			\
	}

	PROCESS_KEY(SCANCODE_CURSORBLOCKRIGHT, KEY_RIGHTARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKLEFT, KEY_LEFTARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKDOWN, KEY_DOWNARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKUP, KEY_UPARROW);

	PROCESS_KEY(SCANCODE_ESCAPE, KEY_ESCAPE);
	PROCESS_KEY(SCANCODE_ENTER, KEY_ENTER);
	PROCESS_KEY(SCANCODE_KEYPADENTER, KEY_ENTER);
	PROCESS_KEY(SCANCODE_TAB, KEY_TAB);
	PROCESS_KEY(SCANCODE_SPACE, ' ');
	PROCESS_KEY(SCANCODE_BREAK, KEY_PAUSE);
	PROCESS_KEY(SCANCODE_BREAK_ALTERNATIVE, KEY_PAUSE);

	PROCESS_KEY(SCANCODE_F1, KEY_F1);
	PROCESS_KEY(SCANCODE_F2, KEY_F2);
	PROCESS_KEY(SCANCODE_F3, KEY_F3);
	PROCESS_KEY(SCANCODE_F4, KEY_F4);
	PROCESS_KEY(SCANCODE_F5, KEY_F5);
	PROCESS_KEY(SCANCODE_F6, KEY_F6);
	PROCESS_KEY(SCANCODE_F7, KEY_F7);
	PROCESS_KEY(SCANCODE_F8, KEY_F8);
	PROCESS_KEY(SCANCODE_F9, KEY_F9);
	PROCESS_KEY(SCANCODE_F10, KEY_F10);
	PROCESS_KEY(SCANCODE_F11, KEY_F11);
	PROCESS_KEY(SCANCODE_F12, KEY_F12);
		
	PROCESS_KEY(SCANCODE_1, '1');
	PROCESS_KEY(SCANCODE_2, '2');
	PROCESS_KEY(SCANCODE_3, '3');
	PROCESS_KEY(SCANCODE_4, '4');
	PROCESS_KEY(SCANCODE_5, '5');
	PROCESS_KEY(SCANCODE_6, '6');
	PROCESS_KEY(SCANCODE_7, '7');
	PROCESS_KEY(SCANCODE_8, '8');
	PROCESS_KEY(SCANCODE_9, '9');
	PROCESS_KEY(SCANCODE_0, '0');

	PROCESS_KEY(SCANCODE_MINUS, KEY_MINUS);
	PROCESS_KEY(SCANCODE_KEYPADMINUS, KEY_MINUS);
	PROCESS_KEY(SCANCODE_EQUAL, KEY_EQUALS);
	PROCESS_KEY(SCANCODE_KEYPADPLUS, KEY_EQUALS);

	PROCESS_KEY(SCANCODE_BACKSPACE, KEY_BACKSPACE);

	PROCESS_KEY(SCANCODE_LEFTSHIFT, KEY_RSHIFT);
	PROCESS_KEY(SCANCODE_RIGHTSHIFT, KEY_RSHIFT);
	PROCESS_KEY(SCANCODE_LEFTCONTROL, KEY_RCTRL);
	PROCESS_KEY(SCANCODE_RIGHTCONTROL, KEY_RCTRL);
	PROCESS_KEY(SCANCODE_LEFTALT, KEY_RALT);
	PROCESS_KEY(SCANCODE_RIGHTALT, KEY_RALT);

	PROCESS_KEY(SCANCODE_Q, 'q');
	PROCESS_KEY(SCANCODE_W, 'w');
	PROCESS_KEY(SCANCODE_E, 'e');
	PROCESS_KEY(SCANCODE_R, 'r');
	PROCESS_KEY(SCANCODE_T, 't');
	PROCESS_KEY(SCANCODE_Y, 'y');
	PROCESS_KEY(SCANCODE_U, 'u');
	PROCESS_KEY(SCANCODE_I, 'i');
	PROCESS_KEY(SCANCODE_O, 'o');
	PROCESS_KEY(SCANCODE_P, 'p');

	PROCESS_KEY(SCANCODE_A, 'a');
	PROCESS_KEY(SCANCODE_S, 's');
	PROCESS_KEY(SCANCODE_D, 'd');
	PROCESS_KEY(SCANCODE_F, 'f');
	PROCESS_KEY(SCANCODE_G, 'g');
	PROCESS_KEY(SCANCODE_H, 'h');
	PROCESS_KEY(SCANCODE_J, 'j');
	PROCESS_KEY(SCANCODE_K, 'k');
	PROCESS_KEY(SCANCODE_L, 'l');

	PROCESS_KEY(SCANCODE_Z, 'z');
	PROCESS_KEY(SCANCODE_X, 'x');
	PROCESS_KEY(SCANCODE_C, 'c');
	PROCESS_KEY(SCANCODE_V, 'v');
	PROCESS_KEY(SCANCODE_B, 'b');
	PROCESS_KEY(SCANCODE_N, 'n');
	PROCESS_KEY(SCANCODE_M, 'm');

	PROCESS_KEY(SCANCODE_INSERT, KEY_INSERT);
	PROCESS_KEY(SCANCODE_HOME, KEY_HOME);
	PROCESS_KEY(SCANCODE_PAGEUP, KEY_PAGEUP);
	PROCESS_KEY(SCANCODE_REMOVE, KEY_DELETE);
	PROCESS_KEY(SCANCODE_END, KEY_END);
	PROCESS_KEY(SCANCODE_PAGEDOWN, KEY_PAGEDOWN);

	PROCESS_KEY(SCANCODE_KEYPADDIVIDE, '/');
	PROCESS_KEY(SCANCODE_KEYPADMULTIPLY, '*');

	memcpy (prev_keystate, keystate, sizeof (prev_keystate));
}

/*  */
/*  I_Update */
/*  */
extern patch_t *PatchINumbers[10];

void I_Update (void)
{
   static int	lasttic;
    int		tics;
    int		i;
	int		c,d,u;

    // draws little dots on the bottom of the screen
    if (devparm)
    {

	i = I_GetTime();
	tics = i - lasttic;
	lasttic = i;

	if (tics>999) tics=999;
	c=tics / 100;
	d=(tics-c*100) / 10;
	u=(tics-c*100-d*10);

	if (PatchINumbers[c])
	{
		V_DrawPatch(320-2-24,0,PatchINumbers[c]);
		V_DrawPatch(320-2-16,0,PatchINumbers[d]);
		V_DrawPatch(320-2-8,0,PatchINumbers[u]);
	}
    }

	/*  Copie image finale + synchro VBL */

	if (dblbuffer) {
		vga_copytoplanar256(svgascreen[fbnum],320,(fbnum*320*240)/4,80,320,240);
		vga_setdisplaystart(fbnum*320*240);
		vga_waitretrace();	

		fbnum =1-fbnum;
		screen=svgascreen[fbnum]+320*20;
		R_InitBuffer(scaledviewwidth,viewheight); /* mettre a jour ylookup */
	} else {
		memcpy(vga_getgraphmem(),screen,SCREENWIDTH*SCREENHEIGHT*pixel_size);
	}
}


/*  */
/*  I_ReadScreen */
/*  */
void I_ReadScreen (byte* scr)
{
    memcpy (scr, screen, SCREENWIDTH*SCREENHEIGHT*pixel_size);
}


/*  */
/*  Palette stuff. */
/*  */

/*  */
/*  I_SetPalette */
/*  */
void I_SetPalette (byte* palette)
{
	int		i;
	lighttable_t	tc=0;
	byte		*palette1=palette;

	switch(pixel_size)
	{
		case 1:
			for(i = 0; i < 256; i++)
			{
				byte r,v,b;

				r = gammatable[usegamma][*palette++]>>2;
				v = gammatable[usegamma][*palette++]>>2;
				b = gammatable[usegamma][*palette++]>>2;

				vga_setpalette(i,r,v,b);
			}
			break;
		case 2:
		case 3:
			for(i = 0; i < 256; i++)
			{
				/* Palettes de lumiere */

				int		r1,v1,b1;
				int		c,y,r,v,b;

				c=orig_colormaps[i]*3;

				r1=gammatable[usegamma][palette[c]];
				v1=gammatable[usegamma][palette[c+1]];
				b1=gammatable[usegamma][palette[c+2]];	

				for(y=0;y<NUMCOLORMAPS;y++)
				{
					r=r1-(y*r1)/(NUMCOLORMAPS-1);
					v=v1-(y*v1)/(NUMCOLORMAPS-1);
					b=b1-(y*b1)/(NUMCOLORMAPS-1);	
					switch (bpp)
					{
						case 15:
							tc = ((r>>3)<<10)|((v>>3)<<5)|(b>>3);
							tc |= tc << 16;
							break;
						case 16:
							tc = ((r>>3)<<11)|((v>>2)<<5)|(b>>3);
							tc |= tc << 16;
							break;
						case 24:
							tc = (r<<16)|(v<<8)|b;
							break;
					}
					colormaps[y*256+i]=tc;			
				}

				/* Invulnerabilite */
	
				colormaps[NUMCOLORMAPS*256+i]=colormaps[orig_colormaps[32*256+i]];

				/* Noir */
	
				colormaps[(NUMCOLORMAPS+1)*256+i]=0;

				/* Palettes normales */

				r=gammatable[usegamma][*palette1++];
				v=gammatable[usegamma][*palette1++];
				b=gammatable[usegamma][*palette1++];

				switch (bpp)
				{
					case 15:
						tc = ((r>>3)<<10)|((v>>3)<<5)|(b>>3);
						tc |= tc << 16;
						break;
					case 16:
						tc = ((r>>3)<<11)|((v>>2)<<5)|(b>>3);
						tc |= tc << 16;
						break;
					case 24:
						tc = (r<<16)|(v<<8)|b;
						break;
				}
				truecolor_palette[i]=tc;
			}
			break;
	}
}

void I_InitGraphics(void)
{
	int		mymode;
	static int	firsttime=1;

	if (!firsttime)
		return;
	firsttime = 0;

	vga_init();

	signal(SIGINT, (void (*)(int)) I_Quit);

	mymode=G320x200x256;
	bpp=8;

	if (video8)
	{
		mymode=G320x240x256;
		dblbuffer=true;
	}


	if (vga_setmode (mymode) < 0)
		I_Error ( "Unable to switch to requested screen mode/resolution:\n");

	keyboard_init();

	if (mymode==G320x240x256) {
		vga_setdisplaystart(1);

		svgascreen[0]=malloc(320*240);
		svgascreen[1]=malloc(320*240);
		screen=svgascreen[1];

		memset(svgascreen[0],0,320*240);
		memset(svgascreen[1],0,320*240);
	} else {
		/* Allocate the screen */
		screen = I_AllocLow (SCREENWIDTH*SCREENHEIGHT*pixel_size);
		if (!screen)
			I_Error("Not enough memory for screens\n");
	}

	switch (bpp)
	{
		case 8:
			R_DrawColumn=R_DrawColumn8;
			R_DrawFuzzColumn=
			R_DrawTranslatedFuzzColumn=
			R_DrawAltFuzzColumn=
				R_DrawFuzzColumn8;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn8;
			R_DrawSpan=R_DrawSpan8;
			R_DrawSkyColumn=R_DrawSkyColumn8;

			R_DrawColumnLow=R_DrawColumnLow8;
			R_DrawFuzzColumnLow=
			R_DrawTranslatedFuzzColumnLow=
			R_DrawAltFuzzColumnLow=
				R_DrawFuzzColumnLow8;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow8;
			R_DrawSpanLow=R_DrawSpanLow8;
			R_DrawSkyColumnLow=R_DrawSkyColumnLow8;

			AM_drawFline=AM_drawFline8;
			V_DrawPatch=V_DrawPatch8;
			V_DrawShadowedPatch=V_DrawShadowedPatch8;
			V_DrawFuzzPatch=V_DrawAltFuzzPatch=V_DrawFuzzPatch8;
			pixel_size=1;
			break;
		case 15:
			R_DrawColumn=R_DrawColumn16;
			R_DrawFuzzColumn=
			R_DrawTranslatedFuzzColumn=
			R_DrawAltFuzzColumn=
				R_DrawFuzzColumn16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSkyColumn=R_DrawSkyColumn16;

			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumnLow=
			R_DrawTranslatedFuzzColumnLow=
			R_DrawAltFuzzColumnLow=
				R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpanLow=R_DrawSpanLow16;
			R_DrawSkyColumnLow=R_DrawSkyColumnLow16;

			AM_drawFline=AM_drawFline16;
			V_DrawPatch=V_DrawShadowedPatch=V_DrawPatch16;
			V_DrawFuzzPatch=V_DrawAltFuzzPatch=V_DrawFuzzPatch16;
			pixel_size=2;
			fuzzmask=0x3DEF3DEF;
			break;
		case 16:
			R_DrawColumn=R_DrawColumn16;
			R_DrawFuzzColumn=
			R_DrawTranslatedFuzzColumn=
			R_DrawAltFuzzColumn=
				R_DrawFuzzColumn16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSkyColumn=R_DrawSkyColumn16;

			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumnLow=
			R_DrawTranslatedFuzzColumnLow=
			R_DrawAltFuzzColumnLow=
				R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpanLow=R_DrawSpanLow16;
			R_DrawSkyColumnLow=R_DrawSkyColumnLow16;

			AM_drawFline=AM_drawFline16;
			V_DrawPatch=V_DrawShadowedPatch=V_DrawPatch16;
			V_DrawFuzzPatch=V_DrawAltFuzzPatch=V_DrawFuzzPatch16;
			pixel_size=2;
			fuzzmask=0x7BEF7BEF;
			break;
		case 24:
			R_DrawColumn=R_DrawColumn24;
			R_DrawFuzzColumn=
			R_DrawTranslatedFuzzColumn=
			R_DrawAltFuzzColumn=
				R_DrawFuzzColumn24;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn24;
			R_DrawSpan=R_DrawSpan24;
			R_DrawSkyColumn=R_DrawSkyColumn24;

			R_DrawColumnLow=R_DrawColumnLow24;
			R_DrawFuzzColumnLow=
			R_DrawTranslatedFuzzColumnLow=
			R_DrawAltFuzzColumnLow=
				R_DrawFuzzColumnLow24;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow24;
			R_DrawSpanLow=R_DrawSpanLow24;
			R_DrawSkyColumnLow=R_DrawSkyColumnLow24;

			AM_drawFline=AM_drawFline24;
			V_DrawPatch=V_DrawShadowedPatch=V_DrawPatch24;
			V_DrawFuzzPatch=V_DrawAltFuzzPatch=V_DrawFuzzPatch24;
			pixel_size=3;
			break;
	}

	/*  Init palette */
	I_SetPalette(W_CacheLumpName("PLAYPAL",PU_CACHE));
}
