
/* ************************************************************************** */
/* ** */
/* ** r_draw.c : Heretic 2 : Raven Software, Corp. */
/* ** */
/* ** $RCSfile: r_draw.c,v $ */
/* ** $Revision: 1.11 $ */
/* ** $Date: 96/01/06 18:37:37 $ */
/* ** $Author: bgokey $ */
/* ** */
/* ************************************************************************** */

#include "h2def.h"
#include "r_local.h"

/*

All drawing to the view buffer is accomplished in this file.  The other refresh
files only know about ccordinates, not the architecture of the frame buffer.

*/

byte *viewimage;
int viewwidth, scaledviewwidth, viewheight, viewwindowx, viewwindowy;
byte *ylookup[MAXHEIGHT];
int columnofs[MAXWIDTH];
/* byte translations[3][256]; // color tables for different players */
byte *tinttable; /*  used for translucent sprites */

extern int pixel_size;

/*
==================
=
= R_DrawColumn
=
= Source is the top of the column to scale
=
==================
*/

lighttable_t	*dc_colormap;
int				dc_x;
int				dc_yl;
int				dc_yh;
fixed_t			dc_iscale;
fixed_t			dc_texturemid;
byte			*dc_source;		/*  first pixel in a column (possibly virtual) */

int				dccount;		/*  just for profiling */

void (*R_DrawColumn) (void);
void (*R_DrawColumnLow) (void);

/*
#define FUZZTABLE	50
#define FUZZOFF	(SCREENWIDTH)
int		fuzzoffset[FUZZTABLE] = {
FUZZOFF,-FUZZOFF,FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,-FUZZOFF,-FUZZOFF,-FUZZOFF,FUZZOFF,-FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,-FUZZOFF,-FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,FUZZOFF,FUZZOFF,-FUZZOFF,FUZZOFF
};
int fuzzpos = 0;
*/

void (*R_DrawFuzzColumn) (void);
void (*R_DrawFuzzColumnLow) (void);

/* ============================================================================ */
/*  */
/*  R_DrawAltFuzzColumn */
/*  */
/* ============================================================================ */

void (*R_DrawAltFuzzColumn) (void);
void (*R_DrawAltFuzzColumnLow) (void);

/*
========================
=
= R_DrawTranslatedColumn
=
========================
*/

byte *dc_translation;
byte *translationtables;

void (*R_DrawTranslatedColumn) (void);
void (*R_DrawTranslatedColumnLow) (void);

/* ============================================================================ */
/*  */
/*  R_DrawTranslatedFuzzColumn */
/*  */
/* ============================================================================ */

void (*R_DrawTranslatedFuzzColumn) (void);
void (*R_DrawTranslatedFuzzColumnLow) (void);

/* -------------------------------------------------------------------------- */
/*  */
/*  PROC R_InitTranslationTables */
/*  */
/* -------------------------------------------------------------------------- */

void R_InitTranslationTables (void)
{
	int i;
	byte *transLump;

	/*  Load tint table */
	tinttable = W_CacheLumpName("TINTTAB", PU_STATIC);

	/*  Allocate translation tables */
	translationtables = Z_Malloc(256*3*(MAXPLAYERS-1)+255, 
		PU_STATIC, 0);
	translationtables = (byte *)(((int)translationtables+255)&~255);

	for(i = 0; i < 3*(MAXPLAYERS-1); i++)
	{
		transLump = W_CacheLumpNum(W_GetNumForName("trantbl0")+i, PU_STATIC);
		memcpy(translationtables+i*256, transLump, 256);
		Z_Free(transLump);
	}
}

/*
================
=
= R_DrawSpan
=
================
*/

int				ds_y;
int				ds_x1;
int				ds_x2;
lighttable_t	*ds_colormap;
fixed_t			ds_xfrac;
fixed_t			ds_yfrac;
fixed_t			ds_xstep;
fixed_t			ds_ystep;
byte			*ds_source;		/*  start of a 64*64 tile image */

int				dscount;		/*  just for profiling */

void (*R_DrawSpan) (void);
void (*R_DrawSpanLow) (void);

void (*R_DrawSkyColumn) (int skyTexture,int offset);
void (*R_DrawSkyColumnLow) (int skyTexture,int offset);

void (*R_DrawSkyColumnDouble) (int skyTexture,int skyTexture2,int offset,int offset2);
void (*R_DrawSkyColumnDoubleLow) (int skyTexture,int skyTexture2,int offset,int offset2);

/*
================
=
= R_InitBuffer
=
=================
*/

void R_InitBuffer (int width, int height)
{
	int		i;
	
	viewwindowx = (SCREENWIDTH-width) >> 1;
	for (i=0 ; i<width ; i++)
		columnofs[i] = (viewwindowx + i)*pixel_size;
	if (width == SCREENWIDTH)
		viewwindowy = 0;
	else
		viewwindowy = (SCREENHEIGHT-SBARHEIGHT-height) >> 1;
	for (i=0 ; i<height ; i++)
		ylookup[i] = screen + (i+viewwindowy)*SCREENWIDTH*pixel_size;
}


/*
==================
=
= R_DrawViewBorder
=
= Draws the border around the view for different size windows
==================
*/

boolean BorderNeedRefresh;

extern lighttable_t truecolor_palette[256];

void R_DrawViewBorder (void)
{
	byte	*src, *dest;
	int		x,y;
	unsigned short *destsrc;
	
	if (scaledviewwidth == SCREENWIDTH)
		return;

	src = W_CacheLumpName("F_022", PU_CACHE);
	dest = screen;
	
	switch(pixel_size)
	{
		case 1:
			for (y=0 ; y<SCREENHEIGHT-SBARHEIGHT ; y++) 
			{ 
				for (x=0 ; x<SCREENWIDTH/64 ; x++) 
				{ 
					memcpy (dest, src+((y&63)<<6), 64); 
					dest += 64; 
				} 

				if (SCREENWIDTH&63)
				{ 
					memcpy (dest, src+((y&63)<<6), SCREENWIDTH&63); 
					dest += (SCREENWIDTH&63); 
				} 
			} 
			break;
		case 2:
			destsrc = (unsigned short *) screen; 
	 
			for (y=0 ; y<SCREENHEIGHT-SBARHEIGHT ; y++) 
			{ 
				unsigned short *dest1;
				byte *src1;

				src1 = src + ((y & 63) << 6);
				dest1 = destsrc;

				for (x=0; x<SCREENWIDTH ; x++)
					*dest1++ = truecolor_palette[ src1[x & 63] ];			
				destsrc += SCREENWIDTH;
			} 
			break;
		case 3:
			for (y=0 ; y<SCREENHEIGHT-SBARHEIGHT ; y++) 
			{ 
				byte	*dest1,*src1;

				src1 = src + ((y & 63) << 6);
				dest1 = dest;

				for (x=0; x<SCREENWIDTH ; x++)
				{
					unsigned long couleur;

					couleur = truecolor_palette[ src1[x & 63] ];

					*dest1++ = couleur;
					*dest1++ = couleur >> 8;
					*dest1++ = couleur >> 16;			
				}
				dest += SCREENWIDTH*3;
			} 
			break;
	}
#if 0
	for (y=0 ; y<SCREENHEIGHT-SBARHEIGHT ; y++)
	{
		for (x=0 ; x<SCREENWIDTH/64 ; x++)
		{
			memcpy (dest, src+((y&63)<<6), 64);
			dest += 64;
		}
		if (SCREENWIDTH&63)
		{
			memcpy (dest, src+((y&63)<<6), SCREENWIDTH&63);
			dest += (SCREENWIDTH&63);
		}
	}
#endif

	for(x=viewwindowx; x < viewwindowx+(viewwidth<<detailshift); x += 16)
	{
		V_DrawPatch(x, viewwindowy-4, W_CacheLumpName("bordt", PU_CACHE));
		V_DrawPatch(x, viewwindowy+viewheight, W_CacheLumpName("bordb", 
			PU_CACHE));
	}
	for(y=viewwindowy; y < viewwindowy+viewheight; y += 16)
	{
		V_DrawPatch(viewwindowx-4, y, W_CacheLumpName("bordl", PU_CACHE));
		V_DrawPatch(viewwindowx+(viewwidth<<detailshift), y, W_CacheLumpName("bordr", 
			PU_CACHE));
	}
	V_DrawPatch(viewwindowx-4, viewwindowy-4, W_CacheLumpName("bordtl", 
		PU_CACHE));
	V_DrawPatch(viewwindowx+(viewwidth<<detailshift), viewwindowy-4, 
		W_CacheLumpName("bordtr", PU_CACHE));
	V_DrawPatch(viewwindowx+(viewwidth<<detailshift), viewwindowy+viewheight, 
		W_CacheLumpName("bordbr", PU_CACHE));
	V_DrawPatch(viewwindowx-4, viewwindowy+viewheight, 
		W_CacheLumpName("bordbl", PU_CACHE));
}

/*
==================
=
= R_DrawTopBorder
=
= Draws the top border around the view for different size windows
==================
*/

boolean BorderTopRefresh;

void R_DrawTopBorder (void)
{
	byte	*src, *dest;
	int		x,y;
	unsigned short *destsrc;
	
	if (scaledviewwidth == SCREENWIDTH)
		return;

/*
	if(shareware)
		src = W_CacheLumpName ("FLOOR04", PU_CACHE);
	else
		src = W_CacheLumpName ("FLAT513", PU_CACHE);
*/
	src = W_CacheLumpName("F_022", PU_CACHE);
	dest = screen;
	
	switch(pixel_size)
	{
		case 1:
			for (y=0 ; y<34 ; y++)
			{
				for (x=0 ; x<SCREENWIDTH/64 ; x++)
				{
					memcpy (dest, src+((y&63)<<6), 64);
					dest += 64;
				}
				if (SCREENWIDTH&63)
				{
					memcpy (dest, src+((y&63)<<6), SCREENWIDTH&63);
					dest += (SCREENWIDTH&63);
				}
			}
			break;
		case 2:
			destsrc = (unsigned short *) screen; 
	 
			for (y=0 ; y<34 ; y++) 
			{ 
				unsigned short *dest1;
				byte *src1;

				src1 = src + ((y & 63) << 6);
				dest1 = destsrc;

				for (x=0; x<SCREENWIDTH ; x++)
					*dest1++ = truecolor_palette[ src1[x & 63] ];			
				destsrc += SCREENWIDTH;
			} 
			break;
		case 3:
			dest=screen;
			for (y=0 ; y<34 ; y++) 
			{ 
				byte	*dest1,*src1;

				src1 = src + ((y & 63) << 6);
				dest1 = dest;

				for (x=0; x<SCREENWIDTH ; x++)
				{
					unsigned long couleur;

					couleur = truecolor_palette[ src1[x & 63] ];

					*dest1++ = couleur;
					*dest1++ = couleur >> 8;
					*dest1++ = couleur >> 16;			
				}
				dest += SCREENWIDTH*3;
			} 
			break;
	}



	if(viewwindowy < 35)
	{
		for(x=viewwindowx; x < viewwindowx+(viewwidth<<detailshift); x += 16)
		{
			V_DrawPatch(x, viewwindowy-4, W_CacheLumpName("bordt", PU_CACHE));
		}
		V_DrawPatch(viewwindowx-4, viewwindowy, W_CacheLumpName("bordl", 
			PU_CACHE));
		V_DrawPatch(viewwindowx+(viewwidth<<detailshift), viewwindowy, 
			W_CacheLumpName("bordr", PU_CACHE));
		V_DrawPatch(viewwindowx-4, viewwindowy+16, W_CacheLumpName("bordl", 
			PU_CACHE));
		V_DrawPatch(viewwindowx+(viewwidth<<detailshift), viewwindowy+16, 
			W_CacheLumpName("bordr", PU_CACHE));

		V_DrawPatch(viewwindowx-4, viewwindowy-4, W_CacheLumpName("bordtl", 
			PU_CACHE));
		V_DrawPatch(viewwindowx+(viewwidth<<detailshift), viewwindowy-4, 
			W_CacheLumpName("bordtr", PU_CACHE));
	}
}
