/*  Emacs style mode select   -*- C++ -*-  */
/* ----------------------------------------------------------------------------- */
/*  */
/*  $Id:$ */
/*  */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  */
/*  $Log:$ */
/*  */
/*  DESCRIPTION: */
/* 	Gamma correction LUT stuff. */
/* 	Functions to draw patches (by post) directly to screen. */
/* 	Functions to blit a block to the screen. */
/*  */
/* ----------------------------------------------------------------------------- */


static const char
rcsid[] = "$Id: v_video.c,v 1.5 1997/02/03 22:45:13 b1 Exp $";


#include "i_system.h"
#include "r_local.h"

#include "h2def.h"
/* #include "doomdata.h" */

#include "i_video.h"

/*  Each screen is [SCREENWIDTH*SCREENHEIGHT];  */
extern	byte	*screen;	

/*  V_DrawPatch */
/*  Masks a column based masked pic to the screen.  */
/*  */
void
V_DrawPatch8
( int		x,
  int		y,
/*   int		scrn, */
  patch_t*	patch ) 
{ 

    int		count;
    int		col; 
    column_t*	column; 
    byte*	desttop;
    byte*	dest;
    byte*	source; 
    int		w; 
	 
    y -= SHORT(patch->topoffset); 
    x -= SHORT(patch->leftoffset); 
#ifdef RANGECHECK 
    if (x<0
	||x+SHORT(patch->width) >SCREENWIDTH
	|| y<0
	|| y+SHORT(patch->height)>SCREENHEIGHT )
/* 	|| (unsigned)scrn>4) */
    {
      fprintf( stderr, "Patch at %d,%d exceeds LFB\n", x,y );
      /*  No I_Error abort - what is up with TNT.WAD? */
      fprintf( stderr, "V_DrawPatch: bad patch (ignored)\n");
      return;
    }
#endif 
 
/*     if (!scrn) */
/* 	V_MarkRect (x, y, SHORT(patch->width), SHORT(patch->height));  */

    col = 0; 
    desttop = screen+y*SCREENWIDTH+x; 
	 
    w = SHORT(patch->width); 

    for ( ; col<w ; x++, col++, desttop++)
    { 
	column = (column_t *)((byte *)patch + LONG(patch->columnofs[col])); 
 
	/*  step through the posts in a column  */
	while (column->topdelta != 0xff ) 
	{ 
	    source = (byte *)column + 3; 
	    dest = desttop + column->topdelta*SCREENWIDTH; 
	    count = column->length; 
			 
	    while (count--) 
	    { 
		*dest = *source++; 
		dest += SCREENWIDTH; 
	    } 
	    column = (column_t *)(  (byte *)column + column->length 
				    + 4 ); 
	} 
    }			 
} 
 
extern byte *tinttable;

void V_DrawFuzzPatch8 (int x, int y, patch_t *patch)
{
	int			count,col;
	column_t	*column;
	byte		*desttop, *dest, *source;
	int			w;
	
	y -= SHORT(patch->topoffset);
	x -= SHORT(patch->leftoffset);

	if (x<0||x+SHORT(patch->width) >SCREENWIDTH || y<0 || y+SHORT(patch->height)>SCREENHEIGHT)
		I_Error ("Bad V_DrawPatch");

	col = 0;
	desttop = screen+y*SCREENWIDTH+x;
	
	w = SHORT(patch->width);
	for ( ; col<w ; x++, col++, desttop++)
	{
		column = (column_t *)((byte *)patch + LONG(patch->columnofs[col]));

/*  step through the posts in a column */
	
		while (column->topdelta != 0xff )
		{
			source = (byte *)column + 3;
			dest = desttop + column->topdelta*SCREENWIDTH;
			count = column->length;
			
			while (count--)
			{
				*dest = tinttable[((*dest)<<8) + *source++];
				dest += SCREENWIDTH;
			}
			column = (column_t *)(  (byte *)column + column->length
+ 4 );
		}
	}
}

/*
==================
=
= V_DrawShadowedPatch
=
= Masks a column based masked pic to the screen.
=
==================
*/

/* extern int *tinttable; */

void V_DrawShadowedPatch8(int x, int y, patch_t *patch)
{
	int			count,col;
	column_t	*column;
	byte		*desttop, *dest, *source;
	byte		*desttop2, *dest2;
	int			w;
	
	y -= SHORT(patch->topoffset);
	x -= SHORT(patch->leftoffset);

	if (x<0||x+SHORT(patch->width) >SCREENWIDTH || y<0 || y+SHORT(patch->height)>SCREENHEIGHT)
		I_Error ("Bad V_DrawPatch");

	col = 0;
	desttop = screen+y*SCREENWIDTH+x;
	desttop2 = screen+(y+2)*SCREENWIDTH+x+2;
	
	w = SHORT(patch->width);
	for ( ; col<w ; x++, col++, desttop++, desttop2++)
	{
		column = (column_t *)((byte *)patch + LONG(patch->columnofs[col]));

/*  step through the posts in a column */
	
		while (column->topdelta != 0xff )
		{
			source = (byte *)column + 3;
			dest = desttop + column->topdelta*SCREENWIDTH;
			dest2 = desttop2 + column->topdelta*SCREENWIDTH;
			count = column->length;
			
			while (count--)
			{
				*dest2 = tinttable[((*dest2)<<8)];
				dest2 += SCREENWIDTH;
				*dest = *source++;
				dest += SCREENWIDTH;

			}
			column = (column_t *)(  (byte *)column + column->length
+ 4 );
		}
	}			
}
